# Define functions to call processor

from typing import Tuple, Union, Dict, List

from PySide2 import QtCore

from sohstationviewer.model.data_type_model import DataTypeModel

from sohstationviewer.view.plotting.plotting_widget.plotting_processor import (
    PlottingChannelProcessor)
from sohstationviewer.view.plotting.plotting_widget.plotting_widget import (
    PlottingWidget)
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.view.util.functions import (
    replace_actual_question_chans, remove_not_found_chans)

from sohstationviewer.controller.util import display_tracking_info
from sohstationviewer.controller.plotting_data import get_title

from sohstationviewer.conf import constants as const

from sohstationviewer.database.extract_data import get_chan_plot_info


class MultiThreadedPlottingWidget(PlottingWidget):
    finished = QtCore.Signal()
    stopped = QtCore.Signal()
    notification = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        PlottingWidget.__init__(self, *args, **kwargs)
        self.data_processors: List[PlottingChannelProcessor] = []
        # pref_order: order of channels to be plotted
        self.pref_order: List[str] = []
        # Only one data processor can run at a time, so it is not a big problem
        #
        self.thread_pool = QtCore.QThreadPool()

        # Flag to indicate whether the data is being plotted for the first time
        # or the plot is being zoomed in.
        self.is_first_plotting = True

        # Used to ensure that the user cannot read a new data set when we are
        # zooming in on the waveform plot. Also to prevent user to use ruler
        # or zoom_markers while plotting
        self.is_working = False

        self.finished.connect(self.stopped)
        self.notification.connect(
            lambda msg: display_tracking_info(self.tracking_box, msg))

    def reset_widget(self):
        """
        Reset the widget in order to plot a new data set.
        """
        self.data_processors = []
        self.is_first_plotting = True
        self.is_working = False

    def init_plot(self,
                  d_obj: DataTypeModel,
                  data_time: List[float],
                  key: Union[str, Tuple[str, str]],
                  start_tm: float, end_tm: float,
                  time_ticks_total: int, is_waveform: bool):
        """
        Initiate plotting with gaps, top time bar

        :param d_obj: object of data
        :param data_time: start and end time of data
        :param key: data set's key
        :param start_tm: requested start time to plot
        :param end_tm: requested end time to plot
        :param time_ticks_total: max number of tick to show on time bar
        :param is_waveform: True if this is a WaveformWidget, otherwise False
        """

        self.key = key
        self.processing_log = []  # [(message, type)]
        self.gap_bar = None
        self.date_mode = self.main_window.date_format.upper()
        self.time_ticks_total = time_ticks_total
        self.min_x = max(data_time[0], start_tm)
        self.max_x = min(data_time[1], end_tm)
        self.plot_total = len(self.plotting_data1) + len(self.plotting_data2)
        cond_total = len(self.plotting_data1)
        name = self.name
        if not is_waveform:
            cond_total += len(self.plotting_data2)
            name += " DATA OR MASS POSITION"
        if cond_total == 0:
            title = f"NO {name} DATA TO DISPLAY."
            self.processing_log.append(
                (f"No {name} data to display.", LogType.INFO))
        else:
            title = get_title(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = const.BOTTOM
        self.plotting_bot_pixel = const.BOTTOM_PX
        self.axes = []
        self.timestamp_bar_top = self.plotting_axes.add_timestamp_bar(0.003)
        self.plotting_axes.set_title(title)

        if cond_total == 0:
            self.draw()
            return False
        else:
            self.plotting_axes.add_gap_bar(d_obj.gaps[key])
            return True

    def create_plotting_channel_processors(
            self, plotting_data: Dict,
            need_db_info: bool = False) -> None:
        """
        Create a data processor for each channel data in the order of
            pref_order. If pref_order isn't given, process in order of
            plotting_data.

        :param plotting_data: dict of data by chan_id
        :param need_db_info: flag to get db info
        """
        chan_order = self.pref_order if self.pref_order \
            else sorted(list(plotting_data.keys()))
        chan_order = replace_actual_question_chans(
            chan_order, list(plotting_data.keys()))
        chan_order = remove_not_found_chans(
            chan_order, list(plotting_data.keys()), self.processing_log)

        not_plot_chans = []
        for chan_id in chan_order:
            if need_db_info:
                chan_db_info = get_chan_plot_info(chan_id,
                                                  self.parent.data_type,
                                                  self.c_mode)
                if (chan_db_info['height'] == 0 or
                        chan_db_info['plotType'] == ''):
                    # not draw
                    not_plot_chans.append(chan_id)
                    continue
                if 'DEFAULT' in chan_db_info['channel']:
                    msg = (f"Channel {chan_id}'s "
                           f"definition can't be found in database. It will"
                           f"be displayed in DEFAULT style.")
                    # TODO: give user a way to add it to DB and leave
                    #  instruction here
                    self.processing_log.append((msg, LogType.WARNING))

                plotting_data[chan_id]['chan_db_info'] = chan_db_info
        if not_plot_chans != []:
            msg = (f"The database settings 'plotType' or 'height' show not to "
                   f"be plotted for the following channels: "
                   f"{', '.join( not_plot_chans)}")
            self.processing_log.append((msg, LogType.WARNING))

        self.move_soh_channels_with_link_to_the_end()

        for chan_id in chan_order:
            if 'chan_db_info' not in plotting_data[chan_id]:
                continue
            channel_processor = PlottingChannelProcessor(
                plotting_data[chan_id], chan_id,
                self.min_x, self.max_x,
                True
            )
            self.data_processors.append(channel_processor)
            channel_processor.finished.connect(self.process_channel)
            channel_processor.stopped.connect(self.has_stopped)

    def move_soh_channels_with_link_to_the_end(self):
        """
        In order to plot a channel (channel A) that is linked with another
        channel (channel B), we need to plot B before we plot A. Because the
        order of the channel in the data is not predetermined, we need to
        manually move A to the end of the data set. This is, of course,
        assuming that channel link is at most one level deep.
        """
        channels_to_move = []
        for channel, chan_data in self.plotting_data1.items():
            try:
                linked_channel = chan_data['chan_db_info']['linkedChan']
                if linked_channel not in ['', 'None', None]:
                    channels_to_move.append(channel)
            except KeyError:
                continue
        for channel in channels_to_move:
            self.plotting_data1[channel] = self.plotting_data1.pop(channel)

    def plot_channels(self, d_obj, key, start_tm, end_tm, time_ticks_total,
                      pref_order=[]):
        """
        Prepare to plot waveform/SOH/mass-position data by creating a data
        processor for each channel, then, run the processors.

        :param d_obj: object of data
        :param key: data set's key
        :param start_tm: requested start time to read
        :param end_tm: requested end time to read
        :param time_ticks_total: max number of tick to show on time bar
        :param pref_order: order of channels to be plotted
        """
        self.pref_order = pref_order
        if not self.is_working:
            self.reset_widget()
            self.is_working = True
            start_msg = f'Plotting {self.name} data...'
            display_tracking_info(self.tracking_box, start_msg, LogType.INFO)
            ret = self.init_plot(d_obj, key, start_tm, end_tm,
                                 time_ticks_total)
            if not ret:
                self.draw()
                self.clean_up()
                self.finished.emit()
                return

            self.create_plotting_channel_processors(self.plotting_data1, True)
            self.create_plotting_channel_processors(self.plotting_data2, True)

            self.process_channel()

    @QtCore.Slot()
    def process_channel(self, channel_data=None, channel_id=None):
        """
        Process a channel data. If channel_id and channel_data is
        not None, remove the first data processor from the list of processor
        and plot the channel channel_id. No matter what, the next channel
        processor is started. If there is no more channel processor, the widget
        goes into the steps of finalizing the plot.

        Semantically, if this method is called with no argument, the processing
        is being started. Otherwise, if it receives two arguments, a data
        processor just finished its work and is signalling the widget to plot
        the processed channel and start processing the next one.

        This method has been designed to only start one processor at once. This
        is a deliberate choice in order to avoid running out of memory. Due to
        how they are designed, channels processors use a lot of memory for
        large data sets. If multiple processors are run at once, memory usage
        of the program becomes untenable. It has been observed that a 7.5 GB
        data set can cause an out of memory error.

        :param channel_data: the data of channel_id
        :param channel_id: the name of the channel to be plotted
        """
        if channel_id is not None:
            self.data_processors.pop(0)
            self.notification.emit(f'Plotting channel {channel_id}...')
            self.plot_single_channel(channel_data, channel_id)
            self.draw()

        try:
            channel_processor = self.data_processors[0]
            self.notification.emit(f'Processing channel '
                                   f'{channel_processor.channel_id}...')
            self.thread_pool.start(channel_processor)
        except IndexError:
            self.notification.emit('Finishing up...')
            self.clean_up()

        except Exception as e:
            print(e)

    def plot_single_channel(self, c_data: Dict, chan_id: str):
        """
        Plot the channel chan_id.

        :param c_data: data of the channel which includes down-sampled
            data in keys 'times' and 'data'
        :param chan_id: name of channel
        """
        pass

    def clean_up(self):
        """
        Clean up after all available channels have been plotted. The
        cleanup steps are as follows.
            Display a finish message
            Reset all internal flags
        """
        if self.is_first_plotting:
            self.done()
            finished_msg = f'{self.name} plot finished'
        else:
            super().set_lim()
            self.draw()
            finished_msg = f'Zooming {self.name} is finished'

        display_tracking_info(self.tracking_box, finished_msg, LogType.INFO)

        self.is_working = False
        self.is_first_plotting = False

    def done(self):
        """
        Finish up the plot after all channels in the data have been plotted.
        Also signal to the main window that the waveform/soh plot is finished
        so that it can update its flags.
        """
        self.axes.append(self.plotting.plot_none())
        self.timestamp_bar_bottom = self.plotting_axes.add_timestamp_bar(
            0.003, top=False)
        super().set_lim(first_time=True)
        self.bottom = self.axes[-1].get_ybound()[0]
        self.ruler = self.plotting_axes.add_ruler(
            self.display_color['time_ruler'])
        self.zoom_marker1 = self.plotting_axes.add_ruler(
            self.display_color['zoom_marker'])
        self.zoom_marker2 = self.plotting_axes.add_ruler(
            self.display_color['zoom_marker'])
        # Set view size fit with the given data
        if self.main_widget.geometry().height() < self.plotting_bot_pixel:
            self.main_widget.setFixedHeight(self.plotting_bot_pixel)
        self.draw()
        self.finished.emit()

    def request_stop(self):
        """
        Request the widget to stop plotting.
        """
        # The currently running data processor will always be the first one in
        # the list.
        self.data_processors[0].request_stop()
        # Because the processors are started one at a time, we only need to
        # delete all the processors from the queue. However, we need to keep
        # the currently running data processor in memory to ensure that its
        # downsampler terminate all running threads gracefully.
        self.data_processors = [self.data_processors[0]]

    @QtCore.Slot()
    def has_stopped(self):
        """
        The slot that is called when the last channel processor has terminated
        all running background threads.
        """
        display_tracking_info(self.tracking_box,
                              f'{self.name} plot stopped', LogType.INFO)
        self.is_working = False
        self.stopped.emit()
